{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Rewriting
	( RewriteRule
	, RewriteRules
	, HeadSymbol(..)
	, applyRewriteRules
	, addRewriteRule
	, createRewriteRuleFromType
	, createRewriteRule
	) where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad
import Control.Monad.Trans
import Data.List as List (map, sort, findIndex, intersperse)
import Data.Maybe

import Syntax.Name
import Syntax.Position
import Syntax.Internal
import Syntax.Values
import Syntax.Pretty

import TypeChecker.Context
import TypeChecker.Abstractions
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Options
import TypeChecker.Errors
import {-# SOURCE #-} TypeChecker.Monad.Context
import {-# SOURCE #-} TypeChecker.UnfoldConstants

import Utils.Names
import Utils.Monads
import Utils.Map as Map hiding (findIndex)
import Utils.Error
import Utils.Debug

{--------------------------------------------------------------------------
    RewriteRule rules
 --------------------------------------------------------------------------}

data RulePat
	= RDef Name [Either RulePat ProjectionValue]
	| RVar Name Int [Either RulePat ProjectionValue]
	| Any

arity :: RulePat -> Int
arity Any	    = 1
arity (RDef _ xs)   = sum $ List.map (either arity (const 0)) xs
arity (RVar _ _ xs) = sum $ List.map (either arity (const 0)) xs

data RewriteRule =
	RewriteRule RulePat Value
		    {- lhs   :: RulePat
		    , rhs   :: Value
		    -}

instance Show RulePat where
    show e = show $ toTerm e
	where
	    toTerm e =
		case e of
		    RDef x es	-> foldl appProj (Def noRange x) es
		    RVar x n es	-> foldl appProj (Var noRange x) es
		    Any		-> Meta noRange DontCare
		where
		    appProj e (Left p)		    = App e (toTerm p)
		    appProj e (Right VFst)	    = Project noRange e (Proj noRange 1)
		    appProj e (Right VSnd)	    = Project noRange e (Proj noRange 2)
		    appProj e (Right (VField x))    = Project noRange e (Field noRange x)

instance Show RewriteRule where
    show (RewriteRule lhs rhs) = show lhs ++ " --> " ++ show rhs

instance Raiseable RulePat where
    raiseByFrom k n Any = Any
    raiseByFrom k n (RVar x m ps)	= RVar x (raiseByFrom k n m) (raiseByFrom k n ps)
    raiseByFrom k n (RDef x ps)	= RDef x $ raiseByFrom k n ps

instance Raiseable RewriteRule where
    raiseByFrom k n (RewriteRule lhs rhs) =
	RewriteRule (raiseByFrom k n lhs)
		    (raiseByFrom k (n + m) rhs)
	where
	    m = arity lhs

instance Instantiate RewriteRule where
    instantiateUnder n (RewriteRule lhs rhs) vs =
	    RewriteRule lhs (instantiateUnder (n + m) rhs vs)
	where
	    m = arity lhs

-- | @Left@ means we didn't rewrite (but might have done unfolding) and
--   @Right@ means that we did rewrite.
matchRule :: RewriteRule -> Value -> TC (Either Value Value)
matchRule (RewriteRule lhs rhs) v =
    do	(v', r) <- matchTop v lhs
	case r of
	    Nothing -> return $ Left v'
	    Just vs -> do
                let newRhs = instantiate rhs vs
                verbose 13 $ liftIO $ putStrLn $
                  "Rewrite rule applied: " ++ show v ++
                  "  ---->  " ++ show newRhs
                return $ Right newRhs
    where
	match :: Value -> RulePat -> TC (Value, Maybe [Value])
	match v p =
	    do	v' <- unfoldConstants v
		matchTop v' p

	-- We can't unfold at the top level (loop)
	matchTop :: Value -> RulePat -> TC (Value, Maybe [Value])
	matchTop v Any = return $ (v, Just [v])
	matchTop v p =
	    do	result <- case (vAppProjView v, p) of
		    (VAPVDef x h es, RDef y ps)
			| x == y && length es == length ps	-> matches (vDef x h) es ps
		    (VAPVCon x h ar es, RDef y ps)
			| x == y && length es == length ps	-> matches (vCon x h ar) es ps
		    (VAPVBId n es, RVar _ m ps)
			| n == m && length es == length ps	-> matches (vBId n) es ps
		    _		-> return (v, Nothing)
                let indicator = if isJust $ snd result then
                                  "matched"
                                 else
                                  "failed to match"
                verbose 14 $ liftIO $ putStrLn $
                  indicator ++ " " ++ show v ++ " against " ++ show p
                return result
	    where
		matches v es ps =
		    do	(vs, rs) <- unzip <$> zipWithM matchP es ps
			let r = concat <$> sequence rs
			return (vAppOrProjs v vs, r)

		matchP (IsApp v) (Left p) =
		    do	(w, vs) <- match v p
			return (IsApp w, vs)
		matchP (IsProj p') (Right p)
		    | p == p'	= return (IsProj p, Just [])
		matchP v p = return (v, Nothing)
		
		vApps = foldl vApp
		vAppOrProjs = foldl appProj
		appProj v (IsProj p) = vProj v p
		appProj v (IsApp w) = vApp v w

{--------------------------------------------------------------------------
    A set of rewrite rules
 --------------------------------------------------------------------------}

type RewriteRules = Map HeadSymbol [RewriteRule]

data HeadSymbol	= HDef Name | HVar Int
    deriving (Eq, Ord)

instance Raiseable HeadSymbol where
    raiseByFrom k n (HDef x) = HDef x
    raiseByFrom k n (HVar m) = HVar $ raiseByFrom k n m

instance (Raiseable k, Raiseable v) => Raiseable (Map k v) where
    raiseByFrom k n =
	mapKeysMonotonic (raiseByFrom k n)
	. Map.map (raiseByFrom k n)

-- | Should this do unfolding? No. But we might wanna unfold the spine if it's a
--   variable at the head. That might not be enough. Maybe it should unfold.
--   Let's try without at first.
--   We need to unfold.
applyRewriteRules :: RewriteRules -> Value -> TC (Either Value Value)
applyRewriteRules rules v =
    case vAppProjView v of
	VAPVDef x _ _
	    | Just rs <- Map.lookup (HDef x) rules  -> apply rs v
	VAPVBId n _
	    | Just rs <- Map.lookup (HVar n) rules  -> apply rs v
	_   -> return $ Left v
    where
	apply []     v = return $ Left v
	apply (r:rs) v =
	    do	r <- matchRule r v
		case r of
		    Left v' -> apply rs v'
		    Right w -> return $ Right w

addRewriteRule :: RewriteRule -> RewriteRules -> RewriteRules
addRewriteRule rule@(RewriteRule lhs _) rules =
    case lhs of
	RDef x _    -> add (HDef x)
	RVar _ n _  -> add (HVar n)
	_	    -> error "impossible: TypeChecker.Rewriting.addRewriteRule"
    where
	add h
	    | Map.member h rules    = Map.adjust (++[rule]) h rules
	    | otherwise		    = Map.insert h [rule] rules

{--------------------------------------------------------------------------
    Creating rewrite rules
 --------------------------------------------------------------------------}

createRewriteRuleFromType :: Type -> TC RewriteRule
createRewriteRuleFromType t = create 0 $ unEl t
    where
	create n v =
	    do	v' <- unfoldConstants v
		case vAppView v' of
		    VAVDef (Name _ "==" _) 1 [_,lhs,rhs]    -> createRewriteRule n lhs rhs
		    VAVPi _ f				    ->
			do  x <- fresh $ absName f
			    extendContext x (Abstract $ error "oops") $
				create (n + 1) (unEl $ absBody f)
		    _				->
			sayWhat (fsep $ pwords $ "the type " ++ show t ++ " cannot be used as a rewrite rule")
			$ failure $ customError $ "it's not on the form (x1:A1,..xn:An) -> lhs == rhs"

createRewriteRule :: Int -> Value -> Value -> TC RewriteRule
createRewriteRule n v w =
    do	w' <- unfoldAllConstants w
	v' <- unfoldAllConstants v
	sayWhat (fsep $ pwords $ "(" ++ show n ++ ") " ++ show v' ++ " --> " ++ show w' ++ " is not a valid rewrite rule") $
	    do	validLHSRHS v' w'
		(lhs, reordering) <- createLHS v'
		unless (valid reordering)
		    $ failure $ customError $ "each universal variable must appear exactly once"
		return $ RewriteRule lhs (reorder reordering w')
    where
	createLHS :: Value -> TC (RulePat, [Int])
	createLHS v =
	    do	v' <- unfoldConstants v
		case vAppProjView v' of
		    VAPVBId m vs
			| m >= n     -> do  (ps,nss) <- unzip <$> mapM createLHSP vs
					    gamma <- getContext
					    let x = [ x | (x,Abstract _) <- gamma ] !! m
					    return (RVar x (m - n) ps, concat nss)
		    VAPVBId m []
			| m < n	     -> return (Any, [m])

		    VAPVDef x _ vs   -> do  (ps,nss) <- unzip <$> mapM createLHSP vs
					    return (RDef x ps, concat nss)

		    VAPVCon x _ _ vs -> do  (ps,nss) <- unzip <$> mapM createLHSP vs
					    return (RDef x ps, concat nss)

		    _	-> failure $ customError $ show v' ++ " cannot appear in a left hand side"

	createLHSP (IsProj p) = return (Right p, [])
	createLHSP (IsApp v)  = do  (p, ns) <- createLHS v
				    return (Left p, ns)

	reorder :: [Int] -> Value -> Value
	reorder ns v =
		instantiateUnder n (raiseBy (length ns) v) $ List.map vBId ns
	    where
		ns' = invert ns
		n   = length ns

	invert :: [Int] -> [Int]
	invert ns = List.map pos [0..n-1]
	    where
		n = length ns
		pos i = j
		    where Just j = findIndex (i==) ns

	valid :: [Int] -> Bool
	valid xs = sort xs == [0..n-1]

validLHSRHS :: Value -> Value -> TC ()
validLHSRHS v w
    | okLHS && okRHS	= return ()
    | otherwise		=
	case (vAppProjView v, vAppProjView w) of
	    (VAPVDef x _ es, VAPVDef x' _ es')
		| x == x'	->
		    do	n <- sum <$> zipWithM valid es es'
			unless (n > 0) err
	    (VAPVBId n es, VAPVBId n' es')
		| n == n'	->
		    do	n <- sum <$> zipWithM valid es es'
			unless (n > 0) err
	    _	-> err
    where
	err :: TC a
	err = do unless okRHS $ failure $ customError $ show w ++ " is not a valid right hand side"
		 unless okLHS $ failure $ customError $ show v ++ " is not a valid left hand side"
		 fail "foo"

	okLHS = validTopLHS v
	okRHS = validRHS w

	valid (IsProj p1) (IsProj p2)
	    | p1 == p2	= return 0
	    | otherwise	= err
	valid (IsApp v) (IsApp w) =
	    do	if v == w then return 0
			  else do validLHSRHS v w
				  return 1
	valid _ _ = err

-- | We allow rewriting to any variable. This means that we have to disallow
--   left hand sides consisting of a single variable (to avoid x --> y, y --> x).
--   This is a reasonable restriction anyway since you can pattern match directly
--   on variables, and if you have an equality proof where the left hand side is
--   a variable you can pattern match on the proof.
--   Otherwise a right hand side must be a fully applied constructor.
validRHS :: Value -> Bool
validRHS v =
    case vAppView v of
	VAVBId _ []	    -> True
	VAVCon _ _ ar vs    -> length vs == ar
	_		    -> False

validTopLHS v =
    case vAppProjView v of
	VAPVBId _ (_:_)	-> True
	VAPVDef _ _ _	-> True
	_		-> False

